function evaluateNetwork

% Required libraries are included
addpath('mmread');
addpath('JPEGtoolbox');

% Name of the sequence to process
sequence = 'akiyo';

% Path of sequence with no compression
pathUncompressed = ['Sequences/' sequence 'Uncompressed/'];
uncompressed = dir(pathUncompressed);

% Load the trained neural network to be used
neuralNetwork = load('trainedNeuralNetwork.mat');
neuralNetwork = neuralNetwork.trainedNeuralNetwork;

% Location of the compressed sequence with MPEG2
pathVideo = ['Sequences/' sequence '.mpg'];
realBitrate = 59.9;     % Specific bitrate of compression

% Location where storing the results
pathSaved = ['Results/' sequence '/'];
directory = [pathSaved '/exec_' datestr(now,30) '/'];

if ~exist(directory,'dir'), mkdir(directory); end

disp(['The aplication of the ARDCT algorithm over the sequence ' sequence ' has started...']);
[files,videoinfo,JPEG_Q,features] = applyNetworkToMPEG2(pathVideo,neuralNetwork,directory,realBitrate);        
disp('...Aplication of the ARDCT completed.');

disp('Obtaining quality measures...');
idx = strfind(files(1).output,'/');
registry.path = files(1).output(1:idx(end)-1);

fileInfo = dir(pathVideo);
tamano = fileInfo.bytes / 1024;
bitrate = realBitrate;

registry.Video = pathVideo;
registry.Tamano = tamano;
registry.Bitrate_Kbits_Seg = bitrate;
registry.Dim = [videoinfo.width videoinfo.height];
registry.Net = neuralNetwork;
registry.JPEG_Q = JPEG_Q;
registry.features = features;
registry.Bitrate_Bits_Pixel = 1000 * (bitrate / (videoinfo.rate*videoinfo.width*videoinfo.height));

registry.MSE_input = zeros(3,length(files));
registry.MSE_output = zeros(3,length(files));
registry.SSIM_input = zeros(3,length(files));
registry.SSIM_output = zeros(3,length(files));

k = 1;
stop = 0;
while k < length(uncompressed) && ~stop
    if (uncompressed(k).isdir) || (isempty(findstr(uncompressed(k).name,'bmp')))
        k = k + 1;
    else
        stop = 1;
    end
end

% Obtaining measures for each bmp file 
for i=1:length(files)
    U = double(imread([pathUncompressed uncompressed(k).name]));
    I = double(imread(files(i).input));
    O = double(imread(files(i).output));

    U_ycc = colorspace('rgb->ycc',uint8(U));
    I_ycc = colorspace('rgb->ycc',uint8(I));
    O_ycc = colorspace('rgb->ycc',uint8(O));
    
    U_ycc_norm(:,:,1) = mat2gray(U_ycc(:,:,1),[16 235]);
    U_ycc_norm(:,:,2) = mat2gray(U_ycc(:,:,2),[16 240]);
    U_ycc_norm(:,:,3) = mat2gray(U_ycc(:,:,3),[16 240]);

    registry.blockiness_U(i) = NormalizedBlockiness(U_ycc_norm);

    I_ycc_norm(:,:,1) = mat2gray(I_ycc(:,:,1),[16 235]);
    I_ycc_norm(:,:,2) = mat2gray(I_ycc(:,:,2),[16 240]);
    I_ycc_norm(:,:,3) = mat2gray(I_ycc(:,:,3),[16 240]);

    registry.blockiness_I(i) = NormalizedBlockiness(I_ycc_norm);

    O_ycc_norm(:,:,1) = mat2gray(O_ycc(:,:,1),[16 235]);
    O_ycc_norm(:,:,2) = mat2gray(O_ycc(:,:,2),[16 240]);
    O_ycc_norm(:,:,3) = mat2gray(O_ycc(:,:,3),[16 240]);

    registry.blockiness_O(i) = NormalizedBlockiness(O_ycc_norm);

    for j=1:3
        registry.MSE_input(j,i) = (sum(sum(((U_ycc(:,:,j) - I_ycc(:,:,j)).^2))))/(size(U_ycc,1)*size(U_ycc,2));
        registry.MSE_output(j,i) = (sum(sum(((U_ycc(:,:,j) - O_ycc(:,:,j)).^2))))/(size(U_ycc,1)*size(U_ycc,2));
        registry.SSIM_input(j,i) = ssim_index(255*U_ycc_norm(:,:,j), 255*I_ycc_norm(:,:,j));
        registry.SSIM_output(j,i) = ssim_index(255*U_ycc_norm(:,:,j), 255*O_ycc_norm(:,:,j));
    end

    idx = strfind(files(i).output,'/');
    registry.file{i} = files(i).output(idx(end)+1:end);
    registry.ID(i) = i;
    k = k + 1;

end

disp('...quality measures obtained.');

R = registry;

%The results are displayed on the screen
MSE_output = (1/3)*mean(R.MSE_output(1,:)) + (1/3)*mean(R.MSE_output(2,:)) + (1/3)*mean(R.MSE_output(3,:));
MSE_input = (1/3)*mean(R.MSE_input(1,:)) + (1/3)*mean(R.MSE_input(2,:)) + (1/3)*mean(R.MSE_input(3,:));

disp(['Average of the MSE for the input sequence: ' num2str(MSE_input)]);
disp(['Average of the MSE for the output sequence (after ARDCT): ' num2str(MSE_output)]);

SSIM_output = (1/3)*mean(R.SSIM_output(1,:)) + (1/3)*mean(R.SSIM_output(2,:)) + (1/3)*mean(R.SSIM_output(3,:));
SSIM_input = (1/3)*mean(R.SSIM_input(1,:)) + (1/3)*mean(R.SSIM_input(2,:)) + (1/3)*mean(R.SSIM_input(3,:));

disp(['Average of the SSIM for the input sequence: ' num2str(SSIM_input)]);
disp(['Average of the SSIM for the output sequence (after ARDCT): ' num2str(SSIM_output)]);

Blockiness_output = mean(R.blockiness_O(1,:));
Blockiness_input = mean(R.blockiness_I(1,:));

disp(['Average of the blockiness for the input sequence: ' num2str(Blockiness_input)]);
disp(['Average of the blockiness for the output sequence (after ARDCT): ' num2str(Blockiness_output)]);

% Store the results
save([registry.path '/' sequence '.mat'],'R');

% Generate a comparative video, where the input stream is on the left and
% the output stream is on the right
disp('Generating a comparative video...');
generateComparativeVideo(registry.path,'bmp');
disp('...Video completed');

% Removing temporal files
delete([registry.path '/*.bmp']);
rmdir([registry.path '/input'],'s');

disp('End of the process');




